
<div class="globaltoc">

<ul>
  <li><a href="{{ pathto('index') }}">index</a></li>
  <li><a href="{{ pathto('install') }}">install</a></li>
  <li><a href="{{ pathto('api') }}">high level API</a></li>
  <li><a href="{{ pathto('lapi') }}">low level API</a></li>
</ul>
<b>external links:</b>
<ul>
  <li><a href="https://github.com/deltachat/deltachat-core-rust">github repository</a></li>
  <li><a href="https://pypi.python.org/pypi/deltachat">pypi: deltachat</a></li>
</ul>

<b>#deltachat [freenode]</b>

</div>
